@@ -3,7 +3,7 @@ import pytest
 from .env import H2Conf, H2TestEnv
 
 
-@pytest.mark.skipif(condition=H2TestEnv.is_unsupported, reason="mod_http2 not supported here")
+@pytest.mark.skipif(condition=H2TestEnv.is_unsupported(), reason="mod_http2 not supported here")
 class TestH2Proxy:
 
     def test_h2_600_01(self, env):
